
/**
 * Title:        JMeter<p>
 * Description:  <p>
 * Copyright:    Copyright (c) Michael Stover<p>
 * Company:      Apache JMeter<p>
 * @author Michael Stover
 * @version 1.0
 */
package org.apache.jmeter.save;

import java.util.*;
import java.io.*;

import org.xml.sax.HandlerBase;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import org.apache.jmeter.util.*;


public class SaveableObjectHandler extends HandlerBase
{
  //Static values
  public final static String CLASS = "class";
  public final static String FILE = "File";
  public final static String FILENAME = "filename";

  ListedHashTree objectHierarchy = new ListedHashTree();
  LinkedList objectStack = new LinkedList();
  HandlerBase currentHandler;
  LinkedList nameStack = new LinkedList();
  LinkedList handlerStack = new LinkedList();
  String errorMessage;
  Saveable firstObject = null;

  int depth = 0;

  public SaveableObjectHandler()
  {
  }

  /**
   * At the end of the document, the hierarchy of saveable objects is traversed,
   * and each saveable object is handed to its "owner".
   */
  public void endDocument()
  {
    System.out.println("End document got called, depth = "+depth);
    if(depth == 0)
    {
      try{
      Object[] top = objectHierarchy.getArray();
      for(int x = 0;x < top.length;x++)
        assembleObject((Saveable)top[x],objectHierarchy.get(top[x]));
      }catch(Exception e){e.printStackTrace();}
    }
  }

  /**
   * Recursive method to add hand all objects to their parent object.  In the
   * end, All the top level objects should contain all the objects below them,
   * even if only indirectly.
   */
  private void assembleObject(Saveable o,ListedHashTree tree)
  {
    Object[] objs;
    if(tree == null || (objs = tree.getArray()).length == 0)
      return;
    else
    {
      for(int x = 0;x < objs.length;x++)
      {
        o.addSaveableObject((Saveable)objs[x]);
        assembleObject((Saveable)objs[x],tree.get(objs[x]));
      }
    }
    return;
  }

  /**
     * org.xml.sax.DocumentHandler implementation.
     **/
    public void characters(char[] data,int start,int length) throws SAXException
    {
      if(currentHandler != null)
        currentHandler.characters(data,start,length);
    }

  /**
   * Gets the objects that were saved into xml.  This assumes the parser has run and
   * is now finished.
   */
  public Saveable[] getSavedObjects()
  {
    Object[] objs = objectHierarchy.getArray();
    Saveable[] ret = new Saveable[objs.length];
    for(int x = 0;x < objs.length;x++)
      ret[x] = (Saveable)objs[x];
    return ret;
  }

  /**
   * If, before you begin, you already have the top level object that you wish to
   * "load" with the saved info, then seed the SaveableObjectHandler here with the
   * object.  Instead of creating a new one, it will use the seeded object.
   */
  public void seedObject(Saveable o)
  {
    firstObject = o;
    depth = 0;
  }

  public void initialize()
  {
    objectHierarchy = new ListedHashTree();
      objectStack = new LinkedList();
      currentHandler = null;
      nameStack = new LinkedList();
      handlerStack = new LinkedList();
  }

  private String resolveClassName(String className)
  {
    if(className != null && className.equals("org.apache.jmeter.controllers.HTTPBaseTestSample"))
      return "org.apache.jmeter.protocol.http.control.HTTPBaseTestSample";
    else return className;
  }

  /**
   * If the element contains information about a java class, then the appropriate
   * object is instantiated and add to the hierarchy of objects.  Stacks are used
   * to record the hierarchy level we're on, as well as providing a simple
   * mechanism to add objects to the hierarchy tree at the right location.
   *
   * If the element is a File element, with a filename attribute, then a new parser
   * is loaded, and that file is read in its entirety, and then back to the regularly
   * scheduled program.  All the objects loaded from the external file get added to
   * the same object hierarchy.
   *
   * Otherwise, the element is simply handed off to the current DocumentHandler in use,
   * for the current object (you didn't expect this handler to handle all possible
   * kinds of xml's, did you?)
   */
  public void startElement(String name,AttributeList atts) throws SAXException
  {
    String className;
    String filename;
    System.out.println("tag name = "+name);
    if((className = atts.getValue(CLASS)) != null)
    {
      className = resolveClassName(className); //temporary fix for package restructuring
      System.out.println("classname = "+className);
      nameStack.addLast(name);
      try {
        Class objClass = Class.forName(className);
        Saveable obj;
        if(firstObject != null)
          obj = firstObject;
        else
          obj = (Saveable)objClass.newInstance();
        firstObject = null;
        objectStack.addLast(obj);
        currentHandler = obj.getHandlerBase();
        currentHandler.startDocument();
        currentHandler.startElement(name,atts);
        System.out.println("currentHandler = "+currentHandler);
        handlerStack.addLast(currentHandler);
        objectHierarchy.add(objectStack,obj);
      }catch(Exception e) {
          errorMessage = "Bad XML in chosen file "+ e.toString();
      }
    }
    else if(name.equals(FILE) && (filename = atts.getValue(FILENAME)) != null)
    {
      System.out.println("filename = "+filename);
      Parser parser = JMeterUtils.getXMLParser();
      parser.setDocumentHandler(this/*new SaveableObjectHandler()*/);
      depth++;
      try {
        parser.parse(new InputSource(new FileInputStream(filename)));
      }catch(Exception e) {
          errorMessage = "Bad XML in chosen file "+ e.toString();
      }
    }
    else if(currentHandler != null)
      currentHandler.startElement(name,atts);
  }

  public void startDocument()
  {}

  /**
   * The element is ended.  If it's an object-creating element, then we can
   * hand the object it's completed DocumentHandler and expect it to
   * configure itself, and then remove stuff from the stack.  Otherwise, pass it
   * on through to the current document handler.
   */
  public void endElement(String name) throws SAXException
  {
    if(nameStack.size() > 0)
    {
    System.out.println("name = "+name+" lastName = "+nameStack.getLast());
      if(name.equals(nameStack.getLast()))
      {
        Saveable obj = (Saveable)objectStack.removeLast();
        System.out.println("Current handler = "+currentHandler);
        currentHandler.endElement(name);
        currentHandler.endDocument();
        try{
        try {
          obj.load(currentHandler);
        }catch(BadHandlerException e) {
            errorMessage = "Bad XML in chosen file "+ e.toString();
        }
        nameStack.removeLast();
        handlerStack.removeLast();
        if(handlerStack.size() > 0)
          currentHandler = (HandlerBase)handlerStack.getLast();
        else currentHandler = null;
        }catch(Exception e){e.printStackTrace();}
      }
    }
    else if(name.equals(FILE))
      depth--;
    else if(currentHandler != null)
      currentHandler.endElement(name);
  }

}
