/*
 * Copyright (c) 1998-99 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Java Apache
 *    Project (http://java.apache.org/)."
 *
 * 4. The names "Apache JMeter" and "Java Apache Project" must
 *    not be used to endorse or promote products derived from this software
 *    without prior written permission.
 *
 * 5. Products derived from this software may not be called "Apache JMeter"
 *    nor may "Java Apache Project" appear in their names without
 *    prior written permission of the Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Java Apache
 *    Project (http://java.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Project. For more information
 * on the Java Apache Project please see <http://java.apache.org/>.
 *
 */

package org.apache.jmeter.util;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.mail.*;
import javax.mail.internet.*;
import org.apache.jmeter.visualizers.*;
import org.xml.sax.*;

/**
 * This class contains the static utility methods used by JMeter.
 *
 * @author  <a href="mailto://stefano@apache.org">Stefano Mazzocchi</a>
 * @version $Revision: 1.1 $ $Date: 2000/11/19 23:31:51 $
 */
public class JMeterUtils {


    private static Properties appProperties;

    /**
     * This method is used by the init method to load the property
     * file that may even reside in the user space, or in the
     * classpath under org.apache.jmeter.jmeter.properties
     */
    public static Properties getProperties(String file) {

        Properties p = new Properties();

        try {
            File f = new File(file);
            p.load(new FileInputStream(f));
        } catch (Exception e) {
            p.clear();
            try {
                p.load(JMeterUtils.class.getClassLoader().getSystemResourceAsStream("org/apache/jmeter/jmeter.properties"));
            } catch (IOException ex) {
                //JMeter.fail("Could not read internal resource. Archive is broken.");
            }
        }

	appProperties = p;
	return p;
    }

    /**
     * This gets the currently defined appProperties.
     * It can only be called after the getProperties(String file)
     * method is called.
     */
    public static Properties getJMeterProperties() {
	return appProperties;
    }

    /**
     * This looks for the requested image in the classpath
     * under org.apache.jmeter.images.<name>
     */
    public static ImageIcon getImage(String name) {
        return new ImageIcon(JMeterUtils.class.getClassLoader().getSystemResource("org/apache/jmeter/images/" + name));
    }

    /**
     * Creates the vector of Timers plugins.
     */
    public static Vector getTimers(Properties properties) {
        return instantiate(getVector(properties, "timer."), "org.apache.jmeter.timers.Timer");
    }

    /**
     * Creates the vector of visualizer plugins.
     */
    public static Vector getVisualizers(Properties properties) {
	return instantiate(getVector(properties, "visualizer."),
			   "org.apache.jmeter.visualizers.Visualizer");
    }

    /**
     * Creates a vector of SampleController plugins.
     * @param properties The properties with information about the samplers
     **/
    public static Vector getControllers(Properties properties) {
      String name = "controller.";
      Vector v = new Vector();
      Enumeration names = properties.keys();
      while (names.hasMoreElements()) {
        String prop = (String) names.nextElement();
        if (prop.startsWith(name)) {
          Object o = instantiate(properties.getProperty(prop),
                  "org.apache.jmeter.control.SamplerController");
          ((org.apache.jmeter.control.SamplerController)o).setPropName(
                  prop.substring(name.length()));
          v.addElement(o);
        }
      }
      return v;
    }

    /**
     * Create a string of class names for a particular SamplerController
     * @param properties The properties with info about the samples.
     * @param name The name of the sampler controller.
     **/
    public static String[] getTestSamples(Properties properties,String name)
    {
      return (String[])getVector(properties, name+".testsample").toArray(new String[0]);
    }

    /**
     * Create an instance of an org.xml.sax.Parser
     * @param properties The properties file containing the parser's class name
     **/
    public static Parser getXMLParser(Properties properties) {
	return (Parser)instantiate(properties.getProperty("xml.parser"),
				   "org.xml.sax.Parser");
    }

    /**
     * Create an instance of an org.xml.sax.Parser based on the default props.
     **/
    public static Parser getXMLParser() {
	return getXMLParser(getJMeterProperties());
    }

    /**
     * Create an instance of the org.xml.sax.DocumentHandler specified.
     * @param properties The properties to use to get the class name.
     * @param value The name of the document handler.
     **/
    public static DocumentHandler getXMLDocumentHandler(Properties properties,
							String value) {
	return (DocumentHandler)instantiate(properties.getProperty
					    ("xml.handler."+value),
					    "org.xml.sax.DocumentHandler");
    }

    /**
     * Create an instance of the org.xml.sax.DocumentHandler specified.
     * @param value The name of the document handler.
     **/
    public static DocumentHandler getXMLDocumentHandler(String value) {
	return getXMLDocumentHandler(getJMeterProperties(), value);
    }

    /**
     * Sets the selection of the JComboBox to the Object 'name' from the
     * list in namVec.
     */
    public static void selJComboBoxItem(Properties properties, JComboBox combo, Vector namVec, String name  ) {
	int idx = namVec.indexOf( name );

	combo.setSelectedIndex( idx );

	// Redisplay.
	combo.updateUI();
	return;
    }

    /**
     * Creates the vector of alias strings.
     */
    public static Hashtable getAlias(Properties properties) {
        return getHashtable(properties, "alias.");
    }

    /**
     * Creates a vector of strings for all the properties that start with a
     * common prefix.
     */
    public static Vector getVector(Properties properties, String name) {
        Vector v = new Vector();
        Enumeration names = properties.keys();
        while (names.hasMoreElements()) {
            String prop = (String) names.nextElement();
            if (prop.startsWith(name)) {
                v.addElement(properties.getProperty(prop));
            }
        }
        return v;
    }

    /**
     * Creates a vector of strings for all the properties that start with a
     * common prefix.
     */
    public static Hashtable getHashtable(Properties properties, String name) {
        Hashtable t = new Hashtable();
        Enumeration names = properties.keys();
        while (names.hasMoreElements()) {
            String prop = (String) names.nextElement();
            if (prop.startsWith(name)) {
                t.put(prop.substring(name.length()), properties.getProperty(prop));
            }
        }
        return t;
    }

    /**
     * Instatiate an object and guarantee its class.
     * @param className The name of the class to instantiate.
     * @param impls The name of the class it subclases.
     **/
    public static Object instantiate(String className, String impls) {
	try {
	    Class c = Class.forName(impls);

	    try {
		Class o = Class.forName(className);

		Object res = o.newInstance();
		if (c.isInstance(res))
		    return res;
		else
		    throw new IllegalArgumentException
			(className+" is not an instance of "+impls);
	    }
	    catch (ClassNotFoundException e) {
		System.err.println("Error loading class " +
				   className + ": class is not found");
	    }
	    catch (IllegalAccessException e) {
		System.err.println("Error loading class " + className +
				   ": does not have access");
	    }
	    catch (InstantiationException e) {
		System.err.println("Error loading class " + className +
				   ": could not instantiate");
	    }
	    catch (NoClassDefFoundError e) {
		System.err.println("Error loading class " + className +
				   ": couldn't find class "+e.getMessage());
	    }
	}
	catch (ClassNotFoundException e) {
	    System.err.println("Error loading class "+impls+
			       ": was not found.");
	}
	return null;
    }

    /**
     * Instantiate a vector of classes
     */
    public static Vector instantiate(Vector v, String className) {
        Vector i = new Vector();

        try {
            Class c = Class.forName(className);
                    Enumeration elements = v.elements();
            while (elements.hasMoreElements()) {
                String name = (String) elements.nextElement();
                try {
                    Object o = Class.forName(name).newInstance();
                    if (c.isInstance(o)) {
                        i.addElement(o);
                    }
                }
		catch (ClassNotFoundException e) {
                    System.err.println("Error loading class " +
				       name + ": class is not found");
                }
		catch (IllegalAccessException e) {
                    System.err.println("Error loading class " + name + ": does not have access");
                }
		catch (InstantiationException e) {
                    System.err.println("Error loading class " + name + ": could not instantiate");
                }
		catch (NoClassDefFoundError e) {
                    System.err.println("Error loading class " + name + ": couldn't find class "+e.getMessage());
                }

            }
        } catch (ClassNotFoundException e) {
            System.err.println("Error loading class " + className + ": class is not found");
        }

        return i;
    }

    /**
     * Tokenize a string into a vector of tokens
     */
    public static Vector tokenize(String string, String separator) {
        Vector v = new Vector();
        StringTokenizer s = new StringTokenizer(string, separator);
        while (s.hasMoreTokens()) {
            v.addElement(s.nextToken());
        }
        return v;
    }

    /**
     * Create a button with the netscape style
     */
    public static JButton createButton(String name, ActionListener listener) {
        JButton button = new JButton(getImage(name + ".on.gif"));
        button.setDisabledIcon(getImage(name + ".off.gif"));
        button.setRolloverIcon(getImage(name + ".over.gif"));
        button.setPressedIcon(getImage(name + ".down.gif"));
        button.setActionCommand(name);
        button.addActionListener(listener);
        button.setRolloverEnabled(true);
        button.setFocusPainted(false);
        button.setBorderPainted(false);
        button.setOpaque(false);
        button.setPreferredSize(new Dimension(24, 24));
        return button;
    }

    /**
     * Create a button with the netscape style
     */
    public static JButton createSimpleButton(String name, ActionListener listener) {
        JButton button = new JButton(getImage(name + ".gif"));
        button.setActionCommand(name);
        button.addActionListener(listener);
        button.setFocusPainted(false);
        button.setBorderPainted(false);
        button.setOpaque(false);
        button.setPreferredSize(new Dimension(25, 25));
        return button;
    }

    /**
     * Get a int value with default if not present.
     * @param propName the name of the property.
     * @param defaultVal the default value.
     **/
    public static int getPropDefault( String propName, int defaultVal ) {
	int ans;
	try {
	    ans = (Integer.valueOf(
			   appProperties.getProperty(propName,Integer.toString(defaultVal)))).intValue();
        } catch (Exception e) {
	    ans = defaultVal;
	}
	return ans;
    }

    /**
     * Get a boolean value with default if not present.
     * @param propName the name of the property.
     * @param defaultVal the default value.
     **/
    public static boolean getPropDefault( String propName, boolean defaultVal ) {
	boolean ans;
	try {
	    String strVal = appProperties.getProperty(propName,
						      (new Boolean(defaultVal)).toString());

	    if ( strVal.equalsIgnoreCase("true") || strVal.equalsIgnoreCase("t") )
		ans = true;
	    else
	    if ( strVal.equalsIgnoreCase("false") || strVal.equalsIgnoreCase("f") )
		ans = false;
	    else
		ans = ((Integer.valueOf(strVal)).intValue() == 1 );
        } catch (Exception e) {
	    ans = defaultVal;
	}
	return ans;
    }

    /**
     * Get a long value with default if not present.
     * @param propName the name of the property.
     * @param defaultVal the default value.
     **/
    public static long getPropDefault( String propName, long defaultVal ) {
	long ans;
	try {
	    ans = (Long.valueOf(appProperties.getProperty(propName,Long.toString(defaultVal)))).longValue();
        } catch (Exception e) {
	    ans = defaultVal;
	}
	return ans;
    }

    /**
     * Get a String value with default if not present.
     * @param propName the name of the property.
     * @param defaultVal the default value.
     **/
    public static String getPropDefault( String propName, String defaultVal ) {
	String ans;
	try {
	    ans = appProperties.getProperty(propName, defaultVal);
        } catch (Exception e) {
	    ans = defaultVal;
	}
	return ans;
    }

 /******************************************************
  Takes a String and a tokenizer character, and returns
   a new array of strings of the string split by the tokenizer
   character.
@param splittee String to be split
@param splitChar Character to split the string on
@return Array of all the tokens.
  ******************************************************/
  public static String[] split(String splittee,String splitChar)
  {
  	if(splittee == null || splitChar == null)
  		return new String[0];
    StringTokenizer tokens;
    String temp;
    int spot;
    while((spot=splittee.indexOf(splitChar+splitChar))!=-1)
      splittee=splittee.substring(0,spot+splitChar.length())+
              splittee.substring(spot+2*splitChar.length(),
              splittee.length());
    Vector returns=new Vector();
    int start = 0,length = splittee.length();
    spot = 0;
    while(start < length && (spot = splittee.indexOf(splitChar,start)) > -1) {
      if(spot > 0)
        returns.addElement(splittee.substring(start,spot));
      start = spot+splitChar.length();
    }
    if(start < length)
      returns.add(splittee.substring(start));
    String[] values=new String[returns.size()];
    returns.copyInto(values);
    return values;
  } // End Method

          /******************************************************
     * Takes a String and a tokenizer character, and returns
     a new array of strings of the string split by the tokenizer
     character.
     @param splittee String to be split
     @param splitChar Character to split the string on
     @param def Default value to place between two split chars that have
     nothing between them
     @return Array of all the tokens.
    ******************************************************/
    public static String[] split(String splittee,String splitChar,String def)
    {
  	if(splittee == null || splitChar == null)
	    return new String[0];
	StringTokenizer tokens;
	String temp;
	int spot;
	while((spot=splittee.indexOf(splitChar+splitChar))!=-1)
	    splittee=splittee.substring(0,spot+splitChar.length())+def+
		splittee.substring(spot+1*splitChar.length(),
				   splittee.length());
	Vector returns=new Vector();
        int start = 0,length = splittee.length();
	spot = 0;
	while(start < length && (spot = splittee.indexOf(splitChar,start)) > -1) {
            if(spot > 0)
	      returns.addElement(splittee.substring(start,spot));
            start = spot+splitChar.length();
	}
        if(start < length)
          returns.add(splittee.substring(start));
	String[] values=new String[returns.size()];
	returns.copyInto(values);
	return values;
    }

    /**
     * Report an error through a dialog box.
     * @param errorMsg the error message.
     * @param dialog the parent dialog.
     **/
    public static void reportErrorToUser(String errorMsg, JDialog dialog) {
	final javax.swing.JDialog theDialog = dialog;
	javax.swing.JTextArea msgBox = new javax.swing.JTextArea(errorMsg);
	theDialog.getContentPane().setLayout(new java.awt.FlowLayout());
	theDialog.getContentPane().add(msgBox);
	msgBox.setBackground(theDialog.getBackground());
	javax.swing.JButton ok = new javax.swing.JButton("OK");
	ok.addActionListener(new ActionListener() {
		public void actionPerformed(java.awt.event.ActionEvent e)
		{
		    theDialog.setVisible(false);
		    theDialog.dispose();
		}
	    });
	Dimension d = msgBox.getPreferredSize();
	Dimension c = ok.getPreferredSize();
	theDialog.getContentPane().add(ok);
	theDialog.setSize((int)d.getWidth()+10,
			  (int)d.getHeight()+(int)c.getHeight()+40);
	theDialog.setTitle("Error!");
	theDialog.setVisible(true);
  }

    /**
     * Report an error through a dialog box.
     * @param errorMsg the error message.
     **/
    public static void reportErrorToUser(String errorMsg)
    {
	reportErrorToUser(errorMsg, new javax.swing.JDialog());
    }

    /**
     * Report an error through a dialog box.
     * @param errorMsg the error message.
     * @param parent the parent frame.
     **/
  public static void reportErrorToUser(String errorMsg,Frame parent)
  {
      reportErrorToUser(errorMsg, new javax.swing.JDialog(parent));
  }

    /******************************************************
   Finds a string in an array of strings and returns the
@param array Array of strings.
@param value String to compare to array values.
@return Index of value in array, or -1 if not in array.
  ******************************************************/
  public static int findInArray(String[] array,String value)
  {
    int count=-1;
    int index=-1;
    if(array != null && value != null)
    {
      while(++count<array.length)
      {
        if(array[count] != null && array[count].equals(value))
        {
          index=count;
          break;
        }
      }
    }
    return index;
  }

    /******************************************************
  Takes an array of strings and a tokenizer character, and returns
   a string of all the strings concatenated with the tokenizer string in between each one.
@param splittee Array of Objects to be concatenated.
@param splitChar Object to unsplit the strings with.
@return Array of all the tokens.
  ******************************************************/
  public static String unsplit(Object[] splittee,Object splitChar)
  {
    StringBuffer retVal=new StringBuffer("");
    int count=-1;
    while(++count<splittee.length)
    {
      if(splittee[count] != null)
        retVal.append(splittee[count]);
      if(count+1<splittee.length && splittee[count+1] != null)
        retVal.append(splitChar);
    }
    return retVal.toString();
  } // End Method

  /******************************************************
  Takes an array of strings and a tokenizer character, and returns
   a string of all the strings concatenated with the tokenizer string in between each one.
@param splittee Array of Objects to be concatenated.
@param splitChar Object to unsplit the strings with.
@param def Default value to replace null values in array.
@return Array of all the tokens.
  ******************************************************/
  public static String unsplit(Object[] splittee,Object splitChar,String def)
  {
    StringBuffer retVal=new StringBuffer("");
    int count=-1;
    while(++count<splittee.length)
    {
      if(splittee[count] != null)
        retVal.append(splittee[count]);
      else
        retVal.append(def);
      if(count+1<splittee.length)
        retVal.append(splitChar);
    }
    return retVal.toString();
  } // End Method

	/**
	 * Prompts the user to choose a file from their filesystems for our own
	 * devious uses.  This method maintains the last directory the user
	 * visited before dismissing the dialog.  This does NOT imply they actually
	 * chose a file from that directory, only that they closed the dialog there.
	 * It is the caller's responsibility to check to see if the selected file
	 * is non-null.
	 *@returns The JFileChooser that interacted with the user, after they are
	 * finished using it (accept or otherwise).
	 */
	public static JFileChooser promptToOpenFile() {
		JFileChooser jfc = null;
		if( lastJFCDirectory == null ) {
			String start = JMeterUtils.getPropDefault("user.dir","");
      if(!start.equals(""))
        jfc = new JFileChooser(start);
      else
			  jfc = new JFileChooser();
		} else {
			jfc = new JFileChooser(lastJFCDirectory);
		}
		jfc.showOpenDialog(null);
		lastJFCDirectory = jfc.getCurrentDirectory().getAbsolutePath();
		return jfc;
	}
	/**
	 * Prompts the user to choose a file from their filesystems for our own
	 * devious uses.  This method maintains the last directory the user
	 * visited before dismissing the dialog.  This does NOT imply they actually
	 * chose a file from that directory, only that they closed the dialog there.
	 * It is the caller's responsibility to check to see if the selected file
	 * is non-null.
	 *@returns The JFileChooser that interacted with the user, after they are
	 * finished using it (accept or otherwise).
	 *@see #promptToOpenFile
	 */
	public static JFileChooser promptToSaveFile() {
		JFileChooser jfc = null;
		if( lastJFCDirectory == null ) {
      String start = JMeterUtils.getPropDefault("user.dir","");
      if(!start.equals(""))
        jfc = new JFileChooser(start);
      else
			  jfc = new JFileChooser();
		} else {
			jfc = new JFileChooser(lastJFCDirectory);
		}
		jfc.showSaveDialog(null);
		lastJFCDirectory = jfc.getCurrentDirectory().getAbsolutePath();
		return jfc;
	}
	/** The last directory visited by the user while choosing Files. */
	public static String lastJFCDirectory = null;
}

