/****************************************************************

Copyright (C) 1999  Michael Stover

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
USA.

Michael Stover can be reached via email at mstover1@rochester.rr.com
or via snail mail at 130 Corwin Rd. Rochester, NY 14610

The following exception to this license exists for Lazer Incorporated:
Lazer Inc is excepted from all limitations and requirements stipulated
in the GPL.  Lazer Inc. is the only entity allowed this limitation.
Lazer does have the right to sell this exception, if they choose, but
they cannot grant additional exceptions to any other entities.

******************************************************************/

//Title:        HashTree
//Author:       Michael Stover
//Company:      Lazer inc.


package org.apache.jmeter.util;

import java.io.*;
import java.util.*;

/**********************************************************************
This class is used to create a tree structure of objects.  Each element in the tree
is also a key to the next node down in the tree.  In the ListedHashTree, the order
in which values are added is preserved (not to be confused with SortedHashTree,
which sorts the order of the values using the compare() function).
***********************************************************************/
public class ListedHashTree implements Serializable
{

  public ListedHashTree()
  {
    data = new HashMap();
    order = new LinkedList();
  }

  public ListedHashTree(Object key)
  {
    data = new HashMap();
    order = new LinkedList();
    data.put(key,new ListedHashTree());
    order.add(key);
  }

  public ListedHashTree(Collection keys)
  {
    data = new HashMap();
    order = new LinkedList();
    Iterator it = keys.iterator();
    while(it.hasNext())
    {
      Object temp = it.next();
      data.put(temp,new ListedHashTree());
      order.add(temp);
    }
  }

  public ListedHashTree(Object[] keys)
  {
    data = new HashMap();
    order = new LinkedList();
    for(int x = 0;x < keys.length;x++)
    {
      data.put(keys[x],new ListedHashTree());
      order.add(keys[x]);
    }
  }

/************************************************************
  If the ListedHashTree contains the given object as a key at the top level, then a
  true result is returned, otherwise false.
@param obj Object to be tested as a key.
@return True if the ListedHashTree contains the key, false otherwise.
***************************************************************/
  public boolean containsKey(Object o)
  {
    return data.containsKey(o);
  }

/*****************************************************************
  If the ListedHashTree is empty, true is returned, false otherwise.
@return True if ListedHashTree is empty, false otherwise.
***************************************************************/
  public boolean isEmpty()
  {
    return data.isEmpty();
  }

/************************************************************************
  Sets a key and it's value in the ListedHashTree.  It actually sets up a key, and then
  creates a node for the key and adds the value to the new node, as a key.
@param key Key to be set up.
@param value Value to be set up as a key in the secondary node.
*************************************************************************/
  public void set(Object key,Object value)
  {
    data.put(key,new ListedHashTree(value));
    order.add(key);
  }

/********************************************************************
 Sets a key and it's values in the ListedHashTree.  It sets up a key in the current node, and
 then creates a node for that key, and adds all the values in the array as keys in
 the new node.
@param key Key to be set up.
@param tree ListedHashTree that the key maps to.
**************************************************************************/
  public void set(Object key,ListedHashTree t)
  {
    data.put(key,t);
    order.add(key);
  }

/************************************************************************
Sets a key and it's values in the ListedHashTree.  It sets up a key in the current node, and
then creates a node for that key, and adds all the values in the array as keys in
 the new node.
@param key Key to be set up.
@param values Array of objects to be added as keys in the secondary node.
********************************************************************/
  public void set(Object key,Object[] values)
  {
    data.put(key,new ListedHashTree(values));
    order.add(key);
  }

/********************************************************************
 Sets a key and it's values in the ListedHashTree.  It sets up a key in the current node, and
 then creates a node for that key, and adds all the values in the array as keys in
 the new node.
@param key Key to be set up.
@param values Collection of objects to be added as keys in the secondary node.
**************************************************************************/
  public void set(Object key,Collection values)
  {
    data.put(key,new ListedHashTree(values));
    order.add(key);
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key array as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the array.  Continues recursing in this
  manner until the end of the first array is reached, at which point all the
  values of the second array are added as keys to the bottom-most node.
@param key Array of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(Object[] key,Object[] values)
  {
    List keys = new LinkedList();
    for(int x = 0;x < key.length;keys.add(key[x++]));
    set(keys,values);
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key array as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the array.  Continues recursing in this
  manner until the end of the first array is reached, at which point all the
  values of the Collection of values are added as keys to the bottom-most node.
@param key Array of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(Object[] key,Collection values)
  {
    List keys = new LinkedList();
    for(int x = 0;x < key.length;keys.add(key[x++]));
    set(keys,values);
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key list as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the list.  Continues recursing in this
  manner until the end of the first list is reached, at which point all the
  values of the array of values are added as keys to the bottom-most node.
@param key List of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(List key,Object[] values)
  {
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      set(key.get(0),values);
    else
    {
      Object temp = key.remove(0);
      add(temp,key.get(0));
      get(temp).set(key,values);
    }
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key list as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the list.  Continues recursing in this
  manner until the end of the first list is reached, at which point all the
  values of the Collection of values are added as keys to the bottom-most node.
@param key List of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(List key,Collection values)
  {
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      set(key.get(0),values);
    else
    {
      Object temp = key.remove(0);
      add(temp,key.get(0));
      get(temp).set(key,values);
    }
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key SortedSet as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the SortedSet.  Continues recursing in this
  manner until the end of the first SortedSet is reached, at which point all the
  values of the array of values are added as keys to the bottom-most node.
@param key SortedSet of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(SortedSet key,Object[] values)
  {
    Object temp;
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      set(key.first(),values);
    else
    {
      temp = key.first();
      key.remove(temp);
      add(temp,key.first());
      get(temp).set(key,values);
    }
  }

/********************************************************************
  Sets a series of keys into the ListedHashTree.  It sets up the first object in the
  key SortedSet as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the SortedSet.  Continues recursing in this
  manner until the end of the first SortedSet is reached, at which point all the
  values of the Collection of values are added as keys to the bottom-most node.
@param key SortedSet of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void set(SortedSet key,Collection values)
  {
    Object temp;
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      set(key.first(),values);
    else
    {
      temp = key.first();
      key.remove(temp);
      add(temp,key.first());
      get(temp).set(key,values);
    }
  }

/*********************************************************************
Adds an key into the ListedHashTree at the current level.
@param key Key to be added to ListedHashTree.
*****************************************************************/
  public void add(Object key)
  {
    if(!data.containsKey(key))
    {
      data.put(key,new ListedHashTree());
      order.add(key);
    }
  }

/*******************************************************************
Adds a bunch of keys into the ListedHashTree at the current level.
@param keys Array of Keys to be added to ListedHashTree.
*******************************************************************/
  public void add(Object[] keys)
  {
    for(int x = 0;x < keys.length;x++)
      add(keys[x]);
  }

/*******************************************************************
Adds a bunch of keys into the ListedHashTree at the current level.
@param keys Collection of Keys to be added to ListedHashTree.
*******************************************************************/
  public void add(Collection keys)
  {
    Iterator it = keys.iterator();
    while(it.hasNext())
      add(it.next());
  }

/************************************************************************
  Adds a key and it's value in the ListedHashTree.  It actually adds  a key, and then
  creates a node for the key and adds the value to the new node, as a key.
@param key Key to be added.
@param value Value to be added as a key in the secondary node.
*************************************************************************/
  public void add(Object key,Object value)
  {
    if(!data.containsKey(key))
      set(key,value);
    else
      get(key).add(value);
  }

/************************************************************************
Adds a key and it's values in the ListedHashTree.  It Adds a key in the current node, and
then creates a node for that key, and adds all the values in the array as keys in
 the new node.
@param key Key to be added.
@param values Array of objects to be added as keys in the secondary node.
********************************************************************/
  public void add(Object key,Object[] values)
  {
    if(!data.containsKey(key))
      set(key,values);
    else
      get(key).add(values);
  }

/************************************************************************
Adds a key and it's values in the ListedHashTree.  It adds a key in the current node, and
then creates a node for that key, and adds all the values in the Collection as keys in
 the new node.
@param key Key to be added.
@param values Collection of objects to be added as keys in the secondary node.
********************************************************************/
  public void add(Object key,Collection values)
  {
    if(!data.containsKey(key))
      set(key,values);
    else
      get(key).add(values);
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key array as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the array.  Continues recursing in this
  manner until the end of the first array is reached, at which point all the
  values of the second array are added as keys to the bottom-most node.
@param key Array of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(Object[] key,Object[] values)
  {
    List list = new LinkedList();
    for(int x = 0;x < key.length;list.add(key[x++]));
    add(list,values);
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key array as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the array.  Continues recursing in this
  manner until the end of the first array is reached, at which point all the
  values of the second Collection are added as keys to the bottom-most node.
@param key Array of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(Object[] key,Collection values)
  {
    List list = new LinkedList();
    for(int x = 0;x < key.length;list.add(key[x++]));
    add(list,values);
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key List as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the List.  Continues recursing in this
  manner until the end of the first List is reached, at which point all the
  values of the second Array are added as keys to the bottom-most node.
@param key List of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(List key,Object[] values)
  {
    if(key == null || key.size() == 0)
    {
      add(values);
      return;
    }
    else if(key.size() == 1)
      add(key.get(0),values);
    else
    {
      Object temp = key.remove(0);
      add(temp,key.get(0));
      get(temp).add(key,values);
    }
  }

    /********************************************************************
  Adds a series of keys into the SortedListedHashTree.  It sets up the first object in the
  key List as a key in the current node, recurses into the next SortedListedHashTree node through
  that key and adds the second object in the List.  Continues recursing in this
  manner until the end of the first List is reached, at which point all the
  values of the second Collection are added as keys to the bottom-most node.
@param key List of keys to put into SortedListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(List key,Object value)
  {
    if(key == null || key.size() == 0)
    {
      add(value);
      return;
    }
    else if(key.size() == 1)
        add(key.get(0),value);
    else
    {
      Object temp = key.remove(0);
      add(temp,key.get(0));
      get(temp).add(key,value);
    }
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key List as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the List.  Continues recursing in this
  manner until the end of the first List is reached, at which point all the
  values of the second Collection are added as keys to the bottom-most node.
@param key List of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(List key,Collection values)
  {
    if(key == null || key.size() == 0)
    {
      add(values);
      return;
    }
    else if(key.size() == 1)
      add(key.get(0),values);
    else
    {
      Object temp = key.remove(0);
      add(temp,key.get(0));
      get(temp).add(key,values);
    }
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key SortedSet as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the SortedSet.  Continues recursing in this
  manner until the end of the first SortedSet is reached, at which point all the
  values of the second Array are added as keys to the bottom-most node.
@param key SortedSet of keys to put into ListedHashTree.
@param values Array of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(SortedSet key,Object[] values)
  {
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      add(key.first(),values);
    else
    {
      Object temp = key.first();
      key.remove(temp);
      add(temp,key.first());
      get(temp).add(key,values);
    }
  }

/********************************************************************
  Adds a series of keys into the ListedHashTree.  It sets up the first object in the
  key SortedSet as a key in the current node, recurses into the next ListedHashTree node through
  that key and adds the second object in the SortedSet.  Continues recursing in this
  manner until the end of the first SortedSet is reached, at which point all the
  values of the second Collection are added as keys to the bottom-most node.
@param key SortedSet of keys to put into ListedHashTree.
@param values Collection of values to be added as keys to bottom-most node.
********************************************************************/
  public void add(SortedSet key,Collection values)
  {
    if(key.size() == 0)
      return;
    else if(key.size() == 1)
      add(key.first(),values);
    else
    {
      Object temp = key.first();
      key.remove(temp);
      add(temp,key.first());
      get(temp).add(key,values);
    }
  }

/*********************************************************************
Gets the ListedHashTree object mapped to the given key.
@param key Key used to find appropriate ListedHashTree()
********************************************************************/
  public ListedHashTree get(Object key)
  {
    return (ListedHashTree)data.get(key);
  }

/*************************************************************
Gets the ListedHashTree object mapped to the last key in the array by recursing through
the ListedHashTree structure one key at a time.
@param keys Array of keys.
@return ListedHashTree at the end of the recursion.
**************************************************************/
  public ListedHashTree get(Object[] keys)
  {
    List list = new LinkedList();
    for(int x = 0;x < keys.length;list.add(keys[x++]));
    return get(list);
  }

/*************************************************************
Gets the ListedHashTree object mapped to the last key in the List by recursing through
the ListedHashTree structure one key at a time.
@param keys List of keys.
@return ListedHashTree at the end of the recursion.
**************************************************************/
  public ListedHashTree get(List keys)
  {
    if(keys.size() == 1)
      return get(keys.get(0));
    else if(keys.size() > 1)
    {
      Object temp = keys.remove(0);
      return get(temp).get(keys);
    }
    return new ListedHashTree();
  }

/*************************************************************
Gets the ListedHashTree object mapped to the last key in the SortedSet by recursing through
the ListedHashTree structure one key at a time.
@param keys SortedSet of keys.
@return ListedHashTree at the end of the recursion.
**************************************************************/
  public ListedHashTree get(SortedSet keys)
  {
    if(keys.size() == 1)
      return get(keys.first());
    else if(keys.size() > 1)
    {
      Object temp = keys.first();
      keys.remove(temp);
      return get(temp).get(keys);
    }
    return new ListedHashTree();
  }

/****************************************************************
Gets a Set of all keys in the current ListedHashTree node.
@return Set of all keys in this ListedHashTree.
*****************************************************************/
  public List list()
  {
    return order;
  }

/***************************************************************
  Gets a Set of all keys in the ListedHashTree mapped to the given key of the
  current ListedHashTree object (in other words, one level down.
@param key Key used to find ListedHashTree to get list of.
@return Set of all keys in found ListedHashTree.
***************************************************************/
  public List list(Object key)
  {
    ListedHashTree temp = (ListedHashTree)data.get(key);
    if(temp != null)
      return temp.list();
    else
      return null;
  }

  public void remove(Object key)
  {
    data.remove(key);
    order.remove(key);
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
array of keys, and returns the Set of keys of the ListedHashTree object at the
end of the recursion.
@param keys Array of keys used to recurse into ListedHashTree structure.
@return Set of all keys found in end ListedHashTree.
**************************************************************/
  public List list(Object[] keys)
  {
    List list = new LinkedList();
    for(int x = 0;x < keys.length;list.add(keys[x++]));
    return list(list);
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
List of keys, and returns the Set of keys of the ListedHashTree object at the
end of the recursion.
@param keys List of keys used to recurse into ListedHashTree structure.
@return Set of all keys found in end ListedHashTree.
**************************************************************/
  public List list(List keys)
  {
    if(keys.size() == 0)
      return list();
    else if(keys.size() == 1)
      return get(keys.get(0)).list();
    else
    {
      Object temp = keys.remove(0);
      return get(temp).list(keys);
    }
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
SortedSet of keys, and returns the Set of keys of the ListedHashTree object at the
end of the recursion.
@param keys SortedSet of keys used to recurse into ListedHashTree structure.
@return Set of all keys found in end ListedHashTree.
**************************************************************/
  public List list(SortedSet keys)
  {
    if(keys.size() == 0)
      return list();
    else if(keys.size() == 1)
      return get(keys.first()).list();
    else
    {
      Object temp = keys.first();
      keys.remove(temp);
      return get(temp).list(keys);
    }
  }

/****************************************************************
Gets an array of all keys in the current ListedHashTree node.
@return Array of all keys in this ListedHashTree.
*****************************************************************/
  public Object[] getArray()
  {
    return order.toArray();
  }

/***************************************************************
  Gets an array of all keys in the ListedHashTree mapped to the given key of the
  current ListedHashTree object (in other words, one level down.
@param key Key used to find ListedHashTree to get list of.
@return Array of all keys in found ListedHashTree.
***************************************************************/
  public Object[] getArray(Object key)
  {
    return get(key).getArray();
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
array of keys, and returns an array of keys of the ListedHashTree object at the
end of the recursion.
@param keys Array of keys used to recurse into ListedHashTree structure.
@return Array of all keys found in end ListedHashTree.
**************************************************************/
  public Object[] getArray(Object[] keys)
  {
    List list = new LinkedList();
    for(int x = 0;x < keys.length;list.add(keys[x++]));
    return getArray(list);
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
List of keys, and returns an array of keys of the ListedHashTree object at the
end of the recursion.
@param keys List of keys used to recurse into ListedHashTree structure.
@return Array of all keys found in end ListedHashTree.
**************************************************************/
  public Object[] getArray(List keys)
  {
    if(keys.size() == 0)
      return getArray();
    if(keys.size() == 1)
      return getArray(keys.get(0));
    else
    {
      Object temp = keys.remove(0);
      return get(temp).getArray(keys);
    }
  }

/*************************************************************
Recurses down into the ListedHashTree stucture using each subsequent key in the
SortedSet of keys, and returns an array of keys of the ListedHashTree object at the
end of the recursion.
@param keys SortedSet of keys used to recurse into ListedHashTree structure.
@return Array of all keys found in end ListedHashTree.
**************************************************************/
  public Object[] getArray(SortedSet keys)
  {
    if(keys.size() == 0)
      return getArray();
    if(keys.size() == 1)
      return getArray(keys.first());
    else
    {
      Object temp = keys.first();
      keys.remove(temp);
      return get(temp).getArray(keys);
    }
  }

  public int hashCode()
  {
    return data.hashCode()*7 + 3;
  }

  public boolean equals(Object o)
  {
    boolean flag = true;
    if(o instanceof ListedHashTree)
    {
      ListedHashTree oo = (ListedHashTree)o;
      Iterator it = order.iterator();
      while(it.hasNext())
      {
        if(!oo.containsKey(it.next()))
        {
          flag = false;
          break;
        }
      }
      if(flag)
      {
        it = order.iterator();
        while(it.hasNext())
        {
          Object temp = it.next();
          flag = get(temp).equals(oo.get(temp));
          if(!flag)
            break;
        }
      }
    }
    else
      flag = false;
    return flag;
  }

  public Set keySet()
  {
    return data.keySet();
  }

/*****************************************************
Searches the ListedHashTree structure for the given key.  If it finds the key,
it returns the ListedHashTree mapped to the key.  If it finds nothing, it returns
null.
@param key Key to search for.
@return ListedHashTree mapped to key, if found, otherwise <B>null</B>.
**********************************************************/
  public ListedHashTree search(Object key)
  {
    ListedHashTree temp = null;
    if(data.containsKey(key))
      temp = (ListedHashTree)data.get(key);
    else
    {
      Iterator it = list().iterator();
      while(it.hasNext())
      {
        if(temp == null)
        {
          temp = ((ListedHashTree)it.next()).search(key);
          break;
        }
        else
          break;
      }
    }
    return temp;
  }




  void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException
  {
    ois.defaultReadObject();
  }

  void writeObject(ObjectOutputStream oos) throws IOException
  {
    oos.defaultWriteObject();
  }

  public int size()
  {
    return data.size();
  }

  private java.util.Map data;
  private List order;
}
